@@ -26,7 +26,7 @@
 
 module Graphics.UI.GLUT.Initialization (
    -- * Primary initialization
-   initialize, getArgsAndInitialize,
+   initialize, getArgsAndInitialize, exit,
 
    -- * Initial window geometry
    initialWindowPosition, initialWindowSize,
@@ -42,40 +42,53 @@ module Graphics.UI.GLUT.Initialization (
    RenderingContext(..), renderingContext,
 
    -- * Direct\/indirect rendering
-   DirectRendering(..), directRendering
+   DirectRendering(..), directRendering,
+
+   -- * OpenGL 3.x context support
+   initialContextVersion, ContextFlag(..), initialContextFlags
 ) where
 
-import Data.Bits ( Bits((.|.),(.&.)) )
-import Data.List ( genericLength, intersperse )
+import Control.Monad ( when )
+import Data.Bits ( Bits((.|.),(.&.)), complement )
+import Data.List ( genericLength, intersperse, mapAccumR )
 import Foreign.C.String ( CString, withCString, peekCString )
 import Foreign.C.Types ( CInt, CUInt )
 import Foreign.Marshal.Array ( withArray0, peekArray )
 import Foreign.Marshal.Utils ( with, withMany )
-import Foreign.Ptr ( Ptr, nullPtr )
+import Foreign.Ptr ( Ptr, nullPtr, nullFunPtr )
 import Foreign.Storable ( Storable(..) )
 import System.Environment ( getProgName, getArgs )
 import Graphics.Rendering.OpenGL.GL.CoordTrans ( Position(..), Size(..) )
 import Graphics.Rendering.OpenGL.GL.StateVar (
    GettableStateVar, makeGettableStateVar,
    SettableStateVar, makeSettableStateVar,
-   StateVar, makeStateVar )
+   StateVar, makeStateVar, HasGetter(..), HasSetter(..) )
 import Graphics.UI.GLUT.Constants (
    glut_INIT_WINDOW_X, glut_INIT_WINDOW_Y,
    glut_INIT_WINDOW_WIDTH, glut_INIT_WINDOW_HEIGHT,
-   glut_RGBA, glut_RGB, glut_INDEX, glut_SINGLE, glut_DOUBLE, glut_ACCUM,
-   glut_ALPHA, glut_DEPTH, glut_STENCIL, glut_MULTISAMPLE, glut_STEREO,
-   glut_LUMINANCE, glut_AUX1, glut_AUX2, glut_AUX3, glut_AUX4,
+   glut_RGBA, glut_RGB, glut_INDEX, glut_LUMINANCE,
+   glut_ALPHA, glut_ACCUM, glut_DEPTH, glut_STENCIL,
+   glut_AUX1, glut_AUX2, glut_AUX3, glut_AUX4,
+   glut_SINGLE, glut_DOUBLE, glut_MULTISAMPLE, glut_STEREO,
+   glut_CAPTIONLESS, glut_BORDERLESS,
    glut_INIT_DISPLAY_MODE,
    glut_DISPLAY_MODE_POSSIBLE,
    glut_RENDERING_CONTEXT, glut_CREATE_NEW_CONTEXT, glut_USE_CURRENT_CONTEXT,
    glut_DIRECT_RENDERING,
    glut_FORCE_INDIRECT_CONTEXT, glut_ALLOW_DIRECT_CONTEXT,
-   glut_TRY_DIRECT_CONTEXT, glut_FORCE_DIRECT_CONTEXT )
+   glut_TRY_DIRECT_CONTEXT, glut_FORCE_DIRECT_CONTEXT,
+   glut_INIT_MAJOR_VERSION, glut_INIT_MINOR_VERSION,
+   glut_DEBUG, glut_FORWARD_COMPATIBLE, glut_INIT_FLAGS )
+import Graphics.UI.GLUT.Extensions
 import Graphics.UI.GLUT.QueryUtils ( simpleGet, glutSetOption )
 import Graphics.UI.GLUT.Types ( Relation(..), relationToString )
 
 --------------------------------------------------------------------------------
 
+#include "HsGLUTExt.h"
+
+--------------------------------------------------------------------------------
+
 -- | Given the program name and command line arguments, initialize the GLUT
 -- library and negotiate a session with the window system. During this
 -- process, 'initialize' may cause the termination of the GLUT program with an
@@ -145,6 +158,16 @@ getArgsAndInitialize = do
    nonGLUTArgs <- initialize prog args
    return (prog, nonGLUTArgs)
 
+-----------------------------------------------------------------------------
+
+-- | (/freeglut only/) De-initialize GLUT. After this, one has to use
+-- 'initialize' or 'getArgsAndInitialize' to initialize GLUT again.
+
+exit :: IO ()
+exit = glutExit
+
+EXTENSION_ENTRY(unsafe,"freeglut",glutExit,IO ())
+
 --------------------------------------------------------------------------------
 
 -- | Controls the /initial window position/.  Windows created by
@@ -214,26 +237,56 @@ foreign import CALLCONV unsafe "glutInitWindowSize" glutInitWindowSize ::
 -- with 'initialDisplayMode'.
 
 data DisplayMode
-   = RGBAMode           -- ^ Select an RGBA mode window. This is the default if neither 'RGBAMode' nor 'IndexMode' are specified.
-   | RGBMode            -- ^ An alias for 'RGBAMode'.
-   | IndexMode          -- ^ Select a color index mode window. This overrides 'RGBAMode' if it is also specified.
-   | LuminanceMode      -- ^ Select a window with a \"luminance\" color model. This model provides the functionality of OpenGL\'s
-                        --   RGBA color model, but the green and blue components are not maintained in the frame buffer. Instead
-                        --   each pixel\'s red component is converted to an index between zero and  'Graphics.UI.GLUT.Colormap.numColorMapEntries'
-                        --   and looked up in a per-window color map to determine the color of pixels within the window. The initial
-                        --   colormap of 'LuminanceMode' windows is initialized to be a linear gray ramp, but can be modified with GLUT\'s
-                        --   colormap actions. /Implementation Notes:/ 'LuminanceMode' is not supported on most OpenGL platforms.
-   | WithAlphaComponent -- ^ Select a window with an alpha component to the color buffer(s).
-   | WithAccumBuffer    -- ^ Select a window with an accumulation buffer.
-   | WithDepthBuffer    -- ^ Select a window with a depth buffer.
-   | WithStencilBuffer  -- ^ Select a window with a stencil buffer.
-   | WithAuxBuffers Int -- ^ (/freeglut only/) Select a window with /n/ (1 .. 4) auxiliary buffers. Any /n/ outside the range 1 .. 4 is a fatal error.
-   | SingleBuffered     -- ^ Select a single buffered window. This is the default if neither 'DoubleBuffered' nor 'SingleBuffered' are specified.
-   | DoubleBuffered     -- ^ Select a double buffered window. This overrides 'SingleBuffered' if it is also specified.
-   | Multisampling      -- ^ Select a window with multisampling support. If multisampling is not available, a non-multisampling
-                        --   window will automatically be chosen. Note: both the OpenGL client-side and server-side implementations
-                        --   must support the @GLX_SAMPLE_SGIS@ extension for multisampling to be available.
-   | Stereoscopic       -- ^ Select A Stereo Window.
+   = RGBAMode
+     -- ^ Select an RGBA mode window. This is the default if neither 'RGBAMode'
+     -- nor 'IndexMode' are specified.
+   | RGBMode
+     -- ^ An alias for 'RGBAMode'.
+   | IndexMode
+     -- ^ Select a color index mode window. This overrides 'RGBAMode' if it is
+     -- also specified.
+   | LuminanceMode
+     -- ^ Select a window with a \"luminance\" color model. This model provides
+     -- the functionality of OpenGL\'s RGBA color model, but the green and blue
+     -- components are not maintained in the frame buffer. Instead each pixel\'s
+     -- red component is converted to an index between zero and
+     --  'Graphics.UI.GLUT.Colormap.numColorMapEntries' and looked up in a
+     -- per-window color map to determine the color of pixels within the window.
+     -- The initial colormap of 'LuminanceMode' windows is initialized to be a
+     -- linear gray ramp, but can be modified with GLUT\'s colormap actions.
+     -- /Implementation Notes:/ 'LuminanceMode' is not supported on most OpenGL
+     -- platforms.
+   | WithAlphaComponent
+     -- ^ Select a window with an alpha component to the color buffer(s).
+   | WithAccumBuffer
+     -- ^ Select a window with an accumulation buffer.
+   | WithDepthBuffer
+     -- ^ Select a window with a depth buffer.
+   | WithStencilBuffer
+     -- ^ Select a window with a stencil buffer.
+   | WithAuxBuffers Int
+     -- ^ (/freeglut only/) Select a window with /n/ (1 .. 4) auxiliary buffers.
+     -- Any /n/ outside the range 1 .. 4 is a fatal error.
+   | SingleBuffered
+     -- ^ Select a single buffered window. This is the default if neither
+     -- 'DoubleBuffered' nor 'SingleBuffered' are specified.
+   | DoubleBuffered
+     -- ^ Select a double buffered window. This overrides 'SingleBuffered' if it
+     -- is also specified.
+   | Multisampling
+     -- ^ Select a window with multisampling support. If multisampling is not
+     -- available, a non-multisampling window will automatically be chosen.
+     -- Note: both the OpenGL client-side and server-side implementations must
+     -- support the @GLX_SAMPLE_SGIS@ extension for multisampling to be
+     -- available. Deprecated, use 'WithSamplesPerPixel'.
+   | WithSamplesPerPixel Int
+     -- ^ Select a window with multisampling, using the given samples per pixel.
+   | Stereoscopic
+     -- ^ Select a stereo window.
+   | Captionless
+     -- ^ Select a window without a caption (/freeglut only/).
+   | Borderless
+     -- ^ Select a window without any borders (/freeglut only/).
    deriving ( Eq, Ord, Show )
 
 marshalDisplayMode :: DisplayMode -> CUInt
@@ -241,10 +294,9 @@ marshalDisplayMode m = case m of
    RGBAMode -> glut_RGBA
    RGBMode -> glut_RGB
    IndexMode -> glut_INDEX
-   SingleBuffered -> glut_SINGLE
-   DoubleBuffered -> glut_DOUBLE
-   WithAccumBuffer -> glut_ACCUM
+   LuminanceMode -> glut_LUMINANCE
    WithAlphaComponent -> glut_ALPHA
+   WithAccumBuffer -> glut_ACCUM
    WithDepthBuffer -> glut_DEPTH
    WithStencilBuffer -> glut_STENCIL
    WithAuxBuffers 1 -> glut_AUX1
@@ -253,9 +305,13 @@ marshalDisplayMode m = case m of
    WithAuxBuffers 4 -> glut_AUX4
    WithAuxBuffers n ->
       error ("marshalDisplayMode: illegal number of auxiliary buffers: " ++ show n)
+   SingleBuffered -> glut_SINGLE
+   DoubleBuffered -> glut_DOUBLE
    Multisampling -> glut_MULTISAMPLE
+   WithSamplesPerPixel _ -> error ("marshalDisplayMode: this should not happen")
    Stereoscopic -> glut_STEREO
-   LuminanceMode -> glut_LUMINANCE
+   Captionless -> glut_CAPTIONLESS
+   Borderless -> glut_BORDERLESS
 
 --------------------------------------------------------------------------------
 
@@ -272,19 +328,38 @@ initialDisplayMode :: StateVar [DisplayMode]
 initialDisplayMode = makeStateVar getInitialDisplayMode setInitialDisplayMode
 
 getInitialDisplayMode :: IO [DisplayMode]
-getInitialDisplayMode = simpleGet i2dms glut_INIT_DISPLAY_MODE
+getInitialDisplayMode = do
+   mode <- simpleGet fromIntegral glut_INIT_DISPLAY_MODE
+   let displayModes = i2dms (mode .&. complement glut_MULTISAMPLE)
+   if mode .&. glut_MULTISAMPLE == 0
+      then return displayModes
+      else do
+         n <- get samplesPerPixel
+         return $ WithSamplesPerPixel n : displayModes
 
-i2dms :: CInt -> [DisplayMode]
-i2dms bitfield =
-   [ c | c <- [ RGBAMode, RGBMode, IndexMode, SingleBuffered, DoubleBuffered,
-                WithAccumBuffer, WithAlphaComponent, WithDepthBuffer,
-                WithStencilBuffer, WithAuxBuffers 1, WithAuxBuffers 2,
-                WithAuxBuffers 3, WithAuxBuffers 4, Multisampling, Stereoscopic,
-                LuminanceMode ]
-       , (fromIntegral bitfield .&. marshalDisplayMode c) /= 0 ]
+i2dms :: CUInt -> [DisplayMode]
+i2dms bitfield | IndexMode `elem` modes || LuminanceMode `elem` modes = modes
+               | otherwise = RGBAMode : modes
+   where modes = i2dmsWithoutRGBA bitfield
+
+i2dmsWithoutRGBA :: CUInt -> [DisplayMode]
+i2dmsWithoutRGBA bitfield =
+   [ c | c <- [ IndexMode, LuminanceMode, WithAlphaComponent,
+                WithAccumBuffer, WithDepthBuffer, WithStencilBuffer,
+                WithAuxBuffers 1, WithAuxBuffers 2, WithAuxBuffers 3,
+                WithAuxBuffers 4, SingleBuffered, DoubleBuffered, Multisampling,
+                Stereoscopic, Captionless, Borderless ]
+       , (bitfield .&. marshalDisplayMode c) /= 0 ]
 
 setInitialDisplayMode :: [DisplayMode] -> IO ()
-setInitialDisplayMode = glutInitDisplayMode . toBitfield marshalDisplayMode
+setInitialDisplayMode modes = do
+   let (spps, transformedModes) = mapAccumR handleMultisampling [] modes
+   mapM_ (samplesPerPixel $=) spps
+   glutInitDisplayMode (toBitfield marshalDisplayMode transformedModes)
+
+handleMultisampling :: [Int] -> DisplayMode -> ([Int], DisplayMode)
+handleMultisampling spps (WithSamplesPerPixel spp) = (spp : spps, Multisampling)
+handleMultisampling spps mode                      = (spps, mode)
 
 toBitfield :: (Bits b) => (a -> b) -> [a] -> b
 toBitfield marshal = foldl (.|.) 0 . map marshal
@@ -301,6 +376,31 @@ displayModePossible =
 
 --------------------------------------------------------------------------------
 
+samplesPerPixel :: StateVar Int
+samplesPerPixel = makeStateVar getSamplesPerPixel setSamplesPerPixel
+
+getSamplesPerPixel :: IO Int
+getSamplesPerPixel = do
+   m <- multisamplingSupported
+   if m
+      then simpleGet fromIntegral (fromIntegral glut_MULTISAMPLE)
+      else return defaultSamplesPerPixels
+
+defaultSamplesPerPixels :: Int
+defaultSamplesPerPixels = 4
+
+setSamplesPerPixel :: Int -> IO ()
+setSamplesPerPixel spp = do
+   m <- multisamplingSupported
+   when m $
+      glutSetOption (fromIntegral glut_MULTISAMPLE) (fromIntegral spp)
+
+multisamplingSupported :: IO Bool
+multisamplingSupported = isKnown "glutGetModeValues"
+   where isKnown = fmap (/= nullFunPtr) . getProcAddressInternal
+
+--------------------------------------------------------------------------------
+
 -- | Capabilities for 'initialDisplayCapabilities', most of them are extensions
 -- of the constructors of 'DisplayMode'.
 
@@ -593,3 +693,80 @@ directRendering =
    makeStateVar
       (simpleGet unmarshalDirectRendering glut_DIRECT_RENDERING)
       (glutSetOption glut_DIRECT_RENDERING . marshalDirectRendering)
+
+-----------------------------------------------------------------------------
+
+-- | (/freeglut only/) Controls the API major\/minor version of the OpenGL
+-- context. If a version less than or equal to 2.1 is requested, the context
+-- returned may implement any version no less than that requested and no
+-- greater than 2.1. If version 3.0 is requested, the context returned must
+-- implement exactly version 3.0. Versioning behavior once GL versions beyond
+-- 3.0 are defined will be defined by an amendment to the OpenGL specification
+-- to define dependencies on such GL versions.
+--
+-- 'Graphics.Rendering.OpenGL.GL.StringQueries.glVersion' and
+-- 'Graphics.Rendering.OpenGL.GL.StringQueries.majorMinor' will return the
+-- actual version supported by a context.
+--
+-- The default context version is (1, 0), which will typically return an
+-- OpenGL 2.1 context, if one is available.
+
+initialContextVersion :: StateVar (Int, Int)
+initialContextVersion = makeStateVar getContextVersion setContextVersion
+
+getContextVersion :: IO (Int, Int)
+getContextVersion = do
+   major <- simpleGet fromIntegral glut_INIT_MAJOR_VERSION
+   minor <- simpleGet fromIntegral glut_INIT_MINOR_VERSION
+   return (major, minor)
+
+setContextVersion :: (Int, Int) -> IO ()
+setContextVersion (major, minor) =
+   glutInitContextVersion (fromIntegral major) (fromIntegral minor)
+
+EXTENSION_ENTRY(unsafe,"freeglut",glutInitContextVersion,CInt -> CInt -> IO ())
+
+-----------------------------------------------------------------------------
+
+-- | A flag affecting the rendering context to create, used in conjunction
+-- with 'initialContextFlags'.
+
+data ContextFlag
+   = -- | Debug contexts are intended for use during application development,
+     -- and provide additional runtime checking, validation, and logging
+     -- functionality while possibly incurring performance penalties. The
+     -- additional functionality provided by debug contexts may vary according
+     -- to the implementation. In some cases a debug context may be identical
+     -- to a non-debug context.
+     DebugContext
+   | -- | Forward-compatible contexts are defined only for OpenGL versions 3.0
+     -- and later. They must not support functionality marked as /deprecated/
+     -- by that version of the API, while a non-forward-compatible context must
+     -- support all functionality in that version, deprecated or not.
+     ForwardCompatibleContext
+   deriving ( Eq, Ord, Show )
+
+marshalContextFlag :: ContextFlag -> CInt
+marshalContextFlag x = case x of
+   DebugContext -> glut_DEBUG
+   ForwardCompatibleContext -> glut_FORWARD_COMPATIBLE
+
+-----------------------------------------------------------------------------
+
+-- | (/freeglut only/) Controls the set of flags for the rendering context.
+
+initialContextFlags :: StateVar [ContextFlag]
+initialContextFlags = makeStateVar getContextFlags setContextFlags
+
+getContextFlags :: IO [ContextFlag]
+getContextFlags = simpleGet i2cfs glut_INIT_FLAGS
+
+i2cfs :: CInt -> [ContextFlag]
+i2cfs bitfield =
+   [ c | c <- [ DebugContext, ForwardCompatibleContext ]
+       , (fromIntegral bitfield .&. marshalContextFlag c) /= 0 ]
+
+setContextFlags :: [ContextFlag] -> IO ()
+setContextFlags = glutInitContextFlags . toBitfield marshalContextFlag
+
+EXTENSION_ENTRY(unsafe,"freeglut",glutInitContextFlags,CInt -> IO ())
